from qgis.PyQt.QtCore import (QT_TRANSLATE_NOOP, QVariant)
from qgis.core import (
  QgsProcessingParameterExtent,
  QgsProcessing,
  QgsProcessingParameterCrs,
  QgsWkbTypes,
  QgsProcessingParameterFeatureSink,
  QgsProcessingFeedback,
  QgsFields,
  QgsField,
  QgsFeature,
  QgsGeometry,
  QgsRectangle,
  QgsVectorLayer,
  QgsProcessingParameterEnum,
  QgsProcessingParameterFeatureSource,
  QgsCoordinateReferenceSystem,
  QgsFeatureRequest,
  QgsCoordinateTransform,
  QgsProject
  )
from .fetchabstract import fetchabstract

from qgis import processing
import math
import itertools

class listtengun(fetchabstract):
  
  # UIs
  PARAMETERS = {  
    "FETCH_EXTENT": {
      "ui_func": QgsProcessingParameterExtent,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("listtengun","Extent for fetching data")
      }
    },
    "FETCH_POLYGON": {
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "optional": True,
        "description": QT_TRANSLATE_NOOP("listtengun","Area for fetching data"),
        "types": [QgsProcessing.TypeVectorPolygon]
      }
    },
    "TARGET_CRS": {
      "ui_func": QgsProcessingParameterCrs,
      "ui_args": {
        "optional": True,
        "description": QT_TRANSLATE_NOOP("listtengun","CRS for the output of the data list")
      }
    },
    "PC_DATABASE": {
      "ui_func": QgsProcessingParameterEnum,
      "ui_args": {
        "options": [],
        "description": QT_TRANSLATE_NOOP("fetchtengun","Point cloud database"),
        "allowMultiple": True,
        "defaultValue": 0
      }
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("listtengun","Point-cloud grid")
      }
    }
  }
  
  MESH_GRID = {}
  FIELDS_MESH_GRID = {    
    "X_CENTER": {"TYPE": QVariant.Double},
    "Y_CENTER": {"TYPE": QVariant.Double}
  }
  FIELDS_TENGUN_LIST = {    
    "crs":          {"TYPE": QVariant.String},
    "meshcode":     {"TYPE": QVariant.String},
    "description":  {"TYPE": QVariant.String},
    "url":          {"TYPE": QVariant.String},
    "file_size":    {"TYPE": QVariant.Double}
  }
  
  # initialization of the algorithm
  def initAlgorithm(self, config):    
    self.initUsingCanvas()
    self.PARAMETERS["PC_DATABASE"]["ui_args"]["options"] = [item["description"] for item in self.DBS]
    self.PARAMETERS["PC_DATABASE"]["ui_args"]["defaultValue"] = list(range(len(self.DBS)))
    self.PARAMETERS["TARGET_CRS"]["ui_args"]["defaultValue"] = None
    self.initParameters()  
  
  # set the mesh grid to obtain tengun data
  def setMeshGrid(self, parameters, context, feedback):
    self.MESH_GRID = {}
    
    # initialize the fields of mesh_grid
    mesh_grid_flds = QgsFields()
    for k, v in self.FIELDS_MESH_GRID.items():
      mesh_grid_flds.append(QgsField(k, v["TYPE"]))
    
    # for each database, set mesh_key (name), 
    for db_idx in self.parameterAsEnums(parameters, "PC_DATABASE", context):
      
      # obtain the database information
      db = self.DBS[db_idx]
      pcode_format = db["pcode_format"]
      dx = db["dx"]
      dy = db["dy"]
      crs = db['crs']
      
      
      # get fetch area, using the crs specified in the database
      
      # if the fetch polygon is not given, use the given fetch area. 
      # if else, set the fetch area to the bounding box of the fetch polygon
      fetch_polygon_prj = None
      if self.parameterAsSource(parameters, "FETCH_POLYGON", context) is None:
        fetch_area = self.parameterAsExtent(parameters, "FETCH_EXTENT", context, QgsCoordinateReferenceSystem(crs))
        
      else:
        # obtain a fetch polygon, which is reprojected to the crs of the database
        fetch_polygon = self.parameterAsSource(parameters, "FETCH_POLYGON", context).materialize(QgsFeatureRequest(), feedback)
        fetch_polygon_prj = processing.run(
          "native:reprojectlayer", 
          {
            "INPUT": fetch_polygon,
            "TARGET_CRS": db["crs"],
            "OUTPUT": "TEMPORARY_OUTPUT"
          },
          context = context,
          feedback = feedback,
          is_child_algorithm=True
        )["OUTPUT"]
        
        # obtain the bbox of the fetch polygon
        fetch_polygon_bbox = processing.run(
          "qgis:minimumboundinggeometry", 
          {
            "INPUT": fetch_polygon,
            "TYPE": 0,
            "OUTPUT": "TEMPORARY_OUTPUT"
          },
          context = context,
          feedback = feedback,
          is_child_algorithm=True
        )["OUTPUT"]
        
        fetch_polygon_bbox = context.getMapLayer(fetch_polygon_bbox)
        fetch_area = fetch_polygon_bbox.getFeature(1).geometry().boundingBox()

      
      # set mesh key, which represents the coordinates (incl. crs) of the mesh grid
      mesh_key = f"{dx:.1f}x{dy:.1f}({crs})"
      self.DBS[db_idx]["mesh_key"] = mesh_key
      
      # if the mesh grid is not set, generate the mesh grid
      if mesh_key not in self.MESH_GRID.keys():
        
        # intiialize the mesh grid (QgsVectorLayer)
        vl_init_str = f"polygon?crs={crs}"
        for k, v in self.FIELDS_MESH_GRID.items():
          if v["TYPE"] == QVariant.Double:
            vl_init_str += f"&field={k}:double"
          elif v["TYPE"] == QVariant.Int:
            vl_init_str += f"&field={k}:integer"
        self.MESH_GRID[mesh_key] = QgsVectorLayer(vl_init_str, "mesh_grid", "memory")

        # generate the mesh grid
        meshcode_registered = []
        for col, row in itertools.product(
          range(1 + math.ceil((fetch_area.xMaximum() - fetch_area.xMinimum())/dx)), 
          range(1 + math.ceil((fetch_area.yMaximum() - fetch_area.yMinimum())/dy))
          ):
          x = fetch_area.xMaximum() if fetch_area.xMinimum() + dx * col > fetch_area.xMaximum() else fetch_area.xMinimum() + dx * col
          y = fetch_area.yMaximum() if fetch_area.yMinimum() + dy * row > fetch_area.yMaximum() else fetch_area.yMinimum() + dy * row
          # x = fetch_area.xMinimum() + dx * col
          # y = fetch_area.yMinimum() + dy * row

          # note that the x-y may be out of the range of the specified crs
          try:
            meshcode_str = self.cmptMeshCodeFromXY(x, y, pcode_format=pcode_format)
            if meshcode_str in meshcode_registered:
              continue
            meshcode_registered.append(meshcode_str)
            ft = QgsFeature(mesh_grid_flds)
            ft.setGeometry(self.cmptGeomFromMeshCode(meshcode_str, pcode_format))
            ft["X_CENTER"] = ft.geometry().centroid().asPoint().x()
            ft["Y_CENTER"] = ft.geometry().centroid().asPoint().y()
            self.MESH_GRID[mesh_key].dataProvider().addFeatures([ft])
          except:
            feedback.pushWarning(self.tr("Failed to get a mesh code of ") + db["description"]. self.tr("The extent may be out of the range."))
        
        # if the fetch polygon is given, intersect the fetch polygon and the mesh grid will be registered as the mesh grid
        if fetch_polygon_prj is not None:          
          mesh_grid = processing.run(
            "native:intersection",
            {
              "INPUT": self.MESH_GRID[mesh_key],
              "OVERLAY": fetch_polygon_prj,
              "OUTPUT": "TEMPORARY_OUTPUT"
            },
            context = context,
            feedback = feedback,
            is_child_algorithm = True
          )["OUTPUT"]
          self.MESH_GRID[mesh_key] = context.getMapLayer(mesh_grid)
  
  # set the meta information for obtaining tengun data
  def setWebFetchArgs(self, parameters, context, feedback):
    self.WEBFETCH_ARGS = {
      "REQUEST_ARGS": [],
      "URL_DICT": {},
      "SET": False,
      "DOWNLOADED_FILE": []
    }
    
    for db_idx in self.parameterAsEnums(parameters, "PC_DATABASE", context):
      db = self.DBS[db_idx]
      base_url = db["url"]
      pcode_format = db["pcode_format"]      
    
      if "{pcode}" not in base_url:
        raise Exception(self.tr("URL must contain {pcode}"))
      
      for ft in self.MESH_GRID[db["mesh_key"]].getFeatures():
        x = ft["X_CENTER"]
        y = ft["Y_CENTER"]
        
        try:
          meshcode_str = self.cmptMeshCodeFromXY(x, y, pcode_format=pcode_format)
          
          download_url = base_url.replace("{pcode}", meshcode_str)
          self.WEBFETCH_ARGS["REQUEST_ARGS"].append(
            {
              "method": "GET",
              "url": download_url,
                "crs": db["crs"]
            }
          )
          self.WEBFETCH_ARGS["URL_DICT"].update(
            {
              download_url: {
                "crs": db["crs"],
                "meshcode": meshcode_str, 
                "pcode_format": pcode_format,
                "description": db["description"]
              }
            }
          )
        except Exception as e:
          feedback.reportError(self.tr("Failed to get a mesh code."))
          raise e
        
    
    if self.WEBFETCH_ARGS["REQUEST_ARGS"] is not None:
      self.WEBFETCH_ARGS["SET"] = True
    
  
  # execution of the algorithm
  def processAlgorithm(self, parameters, context, feedback):    
    # import ptvsd
    # ptvsd.debug_this_thread()
        
    # set the mesh grid to obtain tengun data
    self.setMeshGrid(parameters, context, feedback)
    
    # set the meta information for obtaining SRTM data
    self.setWebFetchArgs(parameters, context, feedback)
    
    # download files using the session info
    self.fetchMultipleTenguns(parameters, context, feedback, parallel=True)
    
    
    # intialize the output fields
    output_flds = QgsFields()
    for k, v in self.FIELDS_TENGUN_LIST.items():
      output_flds.append(QgsField(k, v["TYPE"]))
    
    crs = self.parameterAsCrs(parameters, "TARGET_CRS", context)
    if not crs.isValid():
      crs = QgsCoordinateReferenceSystem(self.DBS[self.parameterAsEnums(parameters, "PC_DATABASE", context)[0]]["crs"])
    
    # initialize the sink
    (sink, dest_id) = self.parameterAsSink(
      parameters, "OUTPUT", context, output_flds, QgsWkbTypes.Polygon, crs
    )
    
    # add features to the sink
    for file in self.WEBFETCH_ARGS["DOWNLOADED_FILE"]:
      ft = QgsFeature(output_flds)
      for k, v in file.items():
        if k != "geometry":
          ft[k] = v
      if QgsCoordinateReferenceSystem(file["crs"]).authid() != crs.authid():
        tf = QgsCoordinateTransform(QgsCoordinateReferenceSystem(file["crs"]), crs, QgsProject.instance())
        ft.setGeometry(file["geometry"].transform(tf))        
      else:
        ft.setGeometry(file["geometry"])
        
      sink.addFeature(ft)
      
    return {"OUTPUT": dest_id}
  
  
  # override the method of fetchabstract.py
  # here, only the header of the file is fetched
  def fetchSingleTengun(self, session, request_args:dict, download_msg:str, crs:QgsCoordinateReferenceSystem, feedback_progress: bool, feedback: QgsProcessingFeedback, working_dir:str = None, output_dir:str = None) -> str:
    try:
      response = session.request(stream = True, timeout = 12.0, **request_args)
    except Exception as e:
      feedback.reportError(self.tr("Download was failed."))
      raise e
    # if download is successful
    if response.status_code == 200:
      file_size = int(response.headers.get('content-length', 0))
      
      url_dict = self.WEBFETCH_ARGS["URL_DICT"][request_args["url"]]
      self.WEBFETCH_ARGS["DOWNLOADED_FILE"].append(
        {
          "crs": url_dict["crs"],
          "meshcode": url_dict["meshcode"],
          "url": request_args["url"],
          "file_size": file_size / 1e6,
          "description": url_dict["description"],
          "geometry": self.cmptGeomFromMeshCode(url_dict["meshcode"], url_dict["pcode_format"])
        }
      )
      feedback.pushInfo(request_args["url"] + " " + f"{file_size/1e6:.1f}MB")
      
    
  
  # compute the geometry from the mesh code of 国土基本図図郭
  def cmptGeomFromMeshCode(self, meshcode, pcode_format = {"level": "500","prefix": ""}) -> QgsGeometry:
    meshcode = meshcode.replace(pcode_format["prefix"], "", 1).upper()
    
    dx1 = 40.0e3
    dy1 = 30.0e3
    nx1 = ord(meshcode[1]) - 69 
    ny1 = 74 - ord(meshcode[0])
    x0 = dx1 * nx1
    y0 = dy1 * ny1
    
    if pcode_format["level"] == "50000":
      return QgsGeometry.fromRect(QgsRectangle(x0, y0, x0 + dx1, y0 + dy1))
    
    dx2 = 4.0e3
    dy2 = 3.0e3
    nx2 = int(meshcode[3])
    ny2 = 9 - int(meshcode[2])
    x0 = x0 + dx2 * nx2
    y0 = y0 + dy2 * ny2
    
    if pcode_format["level"] == "5000":
      return QgsGeometry.fromRect(QgsRectangle(x0, y0, x0 + dx2, y0 + dy2))
    
    
    if pcode_format["level"] == "2500":
      dx3 = 2.0e3
      dy3 = 1.5e3
      nx3 = int(meshcode[4]) % 2 - 1
      ny3 = 1 - int(meshcode[4]) // 2
      x0 = x0 + dx3 * nx3
      y0 = y0 + dy3 * ny3  
      return QgsGeometry.fromRect(QgsRectangle(x0, y0, x0 + dx3, y0 + dy3))
    
    elif pcode_format["level"] == "1000":
      dx4 = 800.0
      dy4 = 600.0
      nx4 = ord(meshcode[5]) - 65
      ny4 = 4 - int(meshcode[4])
      x0 = x0 + dx4 * nx4
      y0 = y0 + dy4 * ny4
      return QgsGeometry.fromRect(QgsRectangle(x0, y0, x0 + dx4, y0 + dy4))
    
    else:
      dx5 = 400.0
      dy5 = 300.0
      nx5 = int(meshcode[5])
      ny5 = 9 - int(meshcode[4])
      x0 = x0 + dx5 * nx5
      y0 = y0 + dy5 * ny5
      return QgsGeometry.fromRect(QgsRectangle(x0, y0, x0 + dx5, y0 + dy5))
  
  # Post processing; append layers
  def postProcessAlgorithm(self, context, feedback):
    return {}

  def displayName(self):
    return self.tr("List Point Cloud Properties")

  def createInstance(self):
    return listtengun()
